@@ -1,81 +1,18 @@
+use crate::hashiter::{Iter, IterMut, OwnedIter};
 use crate::util::{allocate, deallocate, round_to_pow2, AllocationKind};
-use crate::{make_hash, MurmurHasher, Value};
-use std::alloc::{Allocator, Global};
-use std::{
+use crate::{make_hash, Value};
+use core::alloc::Allocator;
+use core::{
     borrow::Borrow,
     default::Default,
     hash::{BuildHasher, BuildHasherDefault, Hash, Hasher},
 };
+use std::alloc::Global;
 
-/// The type used for hashed keys.
-pub(crate) type HashedKey = u64;
+/// The default hasher for a [`HashMap`].
+pub(crate) type DefaultHash = std::collections::hash_map::DefaultHasher;
 
-/// Struct which stores a cell in a hash map. A cell is simply a hash (rather
-/// than the key iteself) and the value associated with the key for which the
-/// hash is associated.
-struct Cell<V> {
-    /// The hashed value of the cell.
-    hash: HashedKey,
-    /// The value assosciated with the hash.
-    value: V,
-}
-
-/// Underlying table for a [HashMap]. This stores a pointer to the buckets and
-/// the mask for the number of cells which are stored in the table.
-struct Table<K, V> {
-    /// Pointer to the buckets for the table.
-    buckets: *mut Bucket<K, V>,
-    /// The mask for indexing into the buckets.
-    size_mask: usize,
-}
-
-impl<K, V> Table<K, V> {
-    /// Gets a mutable slice of the table buckets.
-    pub fn bucket_slice_mut(&mut self) -> &mut [Bucket<K, V>] {
-        unsafe { std::slice::from_raw_parts_mut(self.buckets, self.size()) }
-    }
-
-    /// Gets a slice of the table buckets.
-    pub fn bucket_slice(&self) -> &[Bucket<K, V>] {
-        unsafe { std::slice::from_raw_parts(self.buckets, self.size()) }
-    }
-
-    /// Returns the number of cells in the table.
-    pub fn size(&self) -> usize {
-        self.size_mask + 1
-    }
-}
-
-/// Struct which stores buckets of cells. Each bucket stores four cells
-/// and two delta values per cell. The first delta value for a cell provides
-/// the offset to the cell which is the start of the probe chain for the cell.
-/// The second delta value provides the offset to the next link in the probe
-/// chain once a search along the probe chain has started.
-struct Bucket<K, V> {
-    /// Delta values for the cells. The first 4 values are the delta values to
-    /// the start of the probe chain, and the second 4 values are the delta
-    /// values to the next probe in the chain, for each cell, respectively.
-    deltas: [u8; 8],
-    /// Cells for the bucket.
-    cells: [Cell<V>; 4],
-    /// Placeholder for the key
-    _key: std::marker::PhantomData<K>,
-}
-
-/// Defines the result of an insert into the [HashMap].
-enum InsertResult<V> {
-    /// The insertion found a cell for the key, replaced the old value with
-    /// a new one, and returned the old value.
-    Found(V),
-    /// The insertion was performed with a new key, so a new cell was filled.
-    NewInsert,
-    /// No new cell was found for the key withing the linear search range,
-    /// so we overflowed the max delta value and need to move the map to
-    /// a larger table.
-    Overflow(usize),
-}
-
-/// A HashMap implementation which uses a modified form of RobinHood/Hopscotch
+/// A [`HashMap`] implementation which uses a modified form of RobinHood/Hopscotch
 /// probing. This implementation is efficient, roughly 2x the performance of
 /// the default hash map *when using a the same fast hasher*, and a lot more
 /// when using the default hasher. It has also been benchmarked to be as fast/
@@ -94,9 +31,9 @@ enum InsertResult<V> {
 ///
 /// # Threading
 ///
-/// This version is *not* thread-safe. [crate::LeapMap] is a thread-safe version of the
-/// map.
-pub struct HashMap<K, V, H = BuildHasherDefault<MurmurHasher>, A: Allocator = Global> {
+/// This version is *not* thread-safe. [`LeapMap`] is a thread-safe version of
+/// the map.
+pub struct HashMap<K, V, H = BuildHasherDefault<DefaultHash>, A: Allocator = Global> {
     /// Table for the map.
     table: *mut Table<K, V>,
     /// The hasher for the map.
@@ -124,16 +61,21 @@ impl<'a, K, V, H, A: Allocator> HashMap<K, V, H, A> {
     fn get_table_mut(&self) -> &'a mut Table<K, V> {
         unsafe { &mut *self.table }
     }
+
+    /// Returns true if the hash map has been allocated.
+    fn is_allocated(&self) -> bool {
+        !self.table.is_null()
+    }
 }
 
-impl<K, V> HashMap<K, V, BuildHasherDefault<MurmurHasher>, Global>
+impl<K, V> HashMap<K, V, BuildHasherDefault<DefaultHash>, Global>
 where
     K: Eq + Hash + Clone,
     V: Value,
 {
     /// Creates the hash map with space for the default number of elements, which
     /// will use the global allocator for allocation of the map data.
-    pub fn new() -> HashMap<K, V, BuildHasherDefault<MurmurHasher>, Global> {
+    pub fn new() -> HashMap<K, V, BuildHasherDefault<DefaultHash>, Global> {
         Self::new_in(Global)
     }
 
@@ -141,10 +83,10 @@ where
     /// the global allocator for allocation of the map data.
     pub fn with_capacity(
         capacity: usize,
-    ) -> HashMap<K, V, BuildHasherDefault<MurmurHasher>, Global> {
+    ) -> HashMap<K, V, BuildHasherDefault<DefaultHash>, Global> {
         Self::with_capacity_and_hasher_in(
             capacity,
-            BuildHasherDefault::<MurmurHasher>::default(),
+            BuildHasherDefault::<DefaultHash>::default(),
             Global,
         )
     }
@@ -195,6 +137,7 @@ where
         builder: H,
         allocator: A,
     ) -> HashMap<K, V, H, A> {
+        let capacity = round_to_pow2(capacity.max(Self::INITIAL_SIZE));
         let table = Self::allocate_and_init_table(&allocator, capacity);
         HashMap {
             table,
@@ -227,14 +170,19 @@ where
     /// assert_eq!(map.insert(37, 12), None);
     /// assert_eq!(map.insert(37, 14), Some(12));
     /// ```
-    pub fn insert(&mut self, key: K, value: V) -> Option<V> {
+    pub fn insert(&mut self, key: K, value: V) -> Option<V>
+    where
+        K: std::fmt::Debug,
+    {
         let mut state = self.hash_builder.build_hasher();
         key.hash(&mut state);
         let hash = state.finish();
+        debug_assert!(hash != null_hash());
+        println!("{:?} {:?}", key, hash);
         loop {
             let size_mask = self.get_table().size_mask;
             let buckets = self.get_table_mut().bucket_slice_mut();
-            match Self::insert_or_find(hash, value, buckets, size_mask) {
+            match Self::insert_or_find(hash, &key, value, buckets, size_mask) {
                 InsertResult::Overflow(overflow_index) => {
                     // Resize and move into a new map, then try again.
                     // If this happens on the first iteration, then deleted cells
@@ -273,8 +221,8 @@ where
         K: Borrow<Q>,
         Q: Hash + Eq,
     {
-        self.find(make_hash::<K, Q, H>(&self.hash_builder, &key))
-            .map_or(None, |old| if old.is_null() { None } else { Some(old) })
+        self.find(make_hash::<K, Q, H>(&self.hash_builder, key), key)
+            .and_then(|(_k, v)| if v.is_null() { None } else { Some(v) })
     }
 
     /// Returns a mutable reference type to the value corresponding to the `key`.
@@ -299,7 +247,7 @@ where
         let size_mask = table.size_mask;
         let buckets = table.bucket_slice_mut();
         let hash = make_hash::<K, Q, H>(&self.hash_builder, key);
-        Self::find_mut(buckets, hash, size_mask).map_or(None, |old| {
+        Self::find_mut(buckets, hash, key, size_mask).and_then(|old| {
             if old.is_null() {
                 None
             } else {
@@ -308,6 +256,28 @@ where
         })
     }
 
+    /// Returns the key-value pair corresponding to the supplied key.
+    ///
+    /// The supplied key may be any borrower form of the map's key type, but
+    /// [`Hash`] and [`Eq`] on the borrowed form *must* match those for the key
+    /// type.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// let mut map = leapfrog::HashMap::new();
+    /// map.insert(1, 12);
+    /// assert_eq!(map.get_key_value(&1), Some((&1, &12)));
+    /// assert!(map.get(&2).is_none());
+    /// ```
+    pub fn get_key_value<Q: ?Sized>(&self, key: &Q) -> Option<(&K, &V)>
+    where
+        K: Borrow<Q>,
+        Q: Hash + Eq,
+    {
+        self.find(make_hash::<K, Q, H>(&self.hash_builder, key), key)
+    }
+
     /// Returns true if the map contains the specified `key`.
     ///
     /// # Examples
@@ -358,6 +328,40 @@ where
         }
     }
 
+    /// Creates an iterator over a [`HashMap`] which yields immutable key-value
+    /// reference pairs.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use leapfrog::HashMap;
+    ///
+    /// let mut map = HashMap::new();
+    /// map.insert(12, 27);
+    /// assert_eq!(map.iter().count(), 1);
+    /// ```
+    pub fn iter(&'a self) -> Iter<'a, K, V, H, A> {
+        Iter::new(self)
+    }
+
+    /// Creates an iterator over a [`HashMap`] which yields mutable key-value
+    /// reference pairs.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use leapfrog::HashMap;
+    ///
+    /// let mut map = HashMap::new();
+    /// map.insert(12, 27);
+    /// map.iter_mut().for_each(|(_k, v)| *v += 1);
+    /// assert_eq!(map.get(&12), Some(&28));
+    /// assert_eq!(map.iter_mut().count(), 1);
+    /// ```
+    pub fn iter_mut(&'a mut self) -> IterMut<'a, K, V, H, A> {
+        IterMut::new(self)
+    }
+
     /// Returns the length of the map.
     ///
     /// # Examples
@@ -379,6 +383,11 @@ where
         self.size
     }
 
+    /// Returns `true` if the map contains no elements.
+    pub fn is_empty(&self) -> bool {
+        self.len() == 0
+    }
+
     /// Inserts a new cell into the given `buckets`, where the max number of
     /// buckets is `size_mask` + 1.
     ///
@@ -398,6 +407,7 @@ where
     /// with the index of the overflow.
     fn insert_or_find(
         hash: HashedKey,
+        key: &K,
         value: V,
         buckets: &mut [Bucket<K, V>],
         size_mask: usize,
@@ -405,11 +415,12 @@ where
         let mut index = hash as usize;
         {
             let cell = Self::get_cell_mut(buckets, index, size_mask);
-            if cell.hash == hash {
+            if cell.hash == hash && cell.key == *key {
                 let old_value = cell.value;
                 cell.value = value;
                 return InsertResult::Found(old_value);
-            } else if cell.hash == Self::null_hash() {
+            } else if cell.hash == null_hash() {
+                cell.key = key.clone();
                 cell.hash = hash;
                 cell.value = value;
                 return InsertResult::NewInsert;
@@ -421,13 +432,13 @@ where
         // walk along the chain. At each point, we check if the hash is a match;
         // if we find a match, then we have found the cell.
         let mut delta = Self::get_first_delta(buckets, index, size_mask);
-        let first_delta = if delta == 0 { true } else { false };
+        let first_delta = delta == 0;
         while delta != 0 {
             index += delta as usize;
             delta = Self::get_second_delta(buckets, index, size_mask);
 
             let mut cell = Self::get_cell_mut(buckets, index, size_mask);
-            if cell.hash == hash {
+            if hash == cell.hash && cell.key == *key {
                 let old_value = cell.value;
                 cell.value = value;
                 return InsertResult::Found(old_value);
@@ -448,8 +459,9 @@ where
             // We found an empty cell, so reserve it and link it
             // to the previous cell in the same bucket.
             let mut cell = Self::get_cell_mut(buckets, index, size_mask);
-            if cell.hash == Self::null_hash() {
+            if cell.hash == null_hash() {
                 cell.hash = hash;
+                cell.key = key.clone();
                 cell.value = value;
 
                 let offset = (index - prev_link_index) as u8;
@@ -463,7 +475,7 @@ where
 
             // This is single threaded map, so it's impossible for the hash
             // which is a match to appear outside of the probe chain.
-            debug_assert!(cell.hash != hash);
+            debug_assert!(cell.key != *key);
             probes_remaining -= 1;
         }
 
@@ -474,16 +486,19 @@ where
     /// Tries to find the value for the `hash`, without inserting into the map.
     /// This will reuturn a Some(&v) if the find succeeded, otherwise this will
     /// return None.
-    fn find(&self, hash: HashedKey) -> Option<&V> {
-        debug_assert!(hash != Self::null_hash());
+    fn find<Q: ?Sized + Eq>(&self, hash: HashedKey, key: &Q) -> Option<(&K, &V)>
+    where
+        K: Borrow<Q>,
+    {
+        debug_assert!(hash != null_hash());
 
         let buckets = self.get_table().bucket_slice();
         let size_mask = self.get_table().size_mask;
 
         let mut index = hash as usize;
         let cell = Self::get_cell(buckets, index, size_mask);
-        if cell.hash == hash {
-            return Some(&cell.value);
+        if cell.hash == hash && key.eq(cell.key.borrow()) {
+            return Some((&cell.key, &cell.value));
         }
 
         // Now we need to follow the probe chain for the bucket.
@@ -491,8 +506,8 @@ where
         while delta != 0 {
             index += delta as usize;
             let cell = Self::get_cell(buckets, index, size_mask);
-            if cell.hash == hash {
-                return Some(&cell.value);
+            if cell.hash == hash && key.eq(cell.key.borrow()) {
+                return Some((&cell.key, &cell.value));
             }
 
             delta = Self::get_second_delta(buckets, index, size_mask);
@@ -504,14 +519,23 @@ where
     /// Tries to find the value for the `hash` without inserting into the
     /// `buckets`. This will reuturn a reference to the old value if the find
     /// was successful.
-    fn find_mut(buckets: &mut [Bucket<K, V>], hash: HashedKey, size_mask: usize) -> Option<&mut V> {
-        debug_assert!(hash != Self::null_hash());
+    fn find_mut<Q: ?Sized + Eq>(
+        buckets: &'a mut [Bucket<K, V>],
+        hash: HashedKey,
+        key: &Q,
+        size_mask: usize,
+    ) -> Option<&'a mut V>
+    where
+        K: Borrow<Q>,
+    {
+        debug_assert!(hash != null_hash());
 
         let mut index = hash as usize;
         let mut delta = 0u8;
         let mut found = false;
 
-        if Self::get_cell(buckets, index, size_mask).hash == hash {
+        let cell = Self::get_cell(buckets, index, size_mask);
+        if cell.hash == hash && key.eq(cell.key.borrow()) {
             found = true;
         } else {
             delta = Self::get_first_delta(buckets, index, size_mask);
@@ -522,7 +546,8 @@ where
         // something
         while !found && delta != 0 {
             index += delta as usize;
-            if Self::get_cell(buckets, index, size_mask).hash == hash {
+            let cell = Self::get_cell(buckets, index, size_mask);
+            if cell.hash == hash && key.eq(cell.key.borrow()) {
                 found = true;
                 break;
             }
@@ -594,7 +619,13 @@ where
         for source_index in 0..source_size {
             let cell = Self::get_cell(source_buckets, source_index, source_size_mask);
             if !cell.value.is_null() {
-                match Self::insert_or_find(cell.hash, cell.value, dst_buckets, dst_size_mask) {
+                match Self::insert_or_find(
+                    cell.hash,
+                    &cell.key,
+                    cell.value,
+                    dst_buckets,
+                    dst_size_mask,
+                ) {
                     InsertResult::Overflow(_) => {
                         // New bucekts are too small, failed to move.
                         Self::deallocate_table(&self.allocator, dst_table_ptr);
@@ -624,7 +655,7 @@ where
         assert!(cells >= 4 && (cells % 2 == 0));
         let bucket_count = cells >> 2;
         let bucket_ptr =
-            allocate::<Bucket<K, V>, A>(&allocator, bucket_count, AllocationKind::Uninitialized);
+            allocate::<Bucket<K, V>, A>(allocator, bucket_count, AllocationKind::Uninitialized);
         let buckets = unsafe { std::slice::from_raw_parts_mut(bucket_ptr, bucket_count) };
 
         // Since AtomicU8 and AtomicU64 are the same as u8 and u64 in memory,
@@ -636,6 +667,7 @@ where
             };
 
             for cell in 0..4 {
+                // FIXME: How to initialize keys?
                 unsafe {
                     let cell_hash = &mut buckets[i].cells[cell].hash as *mut HashedKey;
                     std::ptr::write_bytes(cell_hash, 0, 1);
@@ -647,7 +679,7 @@ where
             }
         }
 
-        let table_ptr = allocate::<Table<K, V>, A>(&allocator, 1, AllocationKind::Uninitialized);
+        let table_ptr = allocate::<Table<K, V>, A>(allocator, 1, AllocationKind::Uninitialized);
         let table = unsafe { &mut *table_ptr };
 
         table.buckets = bucket_ptr;
@@ -665,9 +697,30 @@ where
         deallocate::<Table<K, V>, A>(allocator, table_ptr, 1);
     }
 
+    pub(crate) fn get_cell_at_index(&self, index: usize) -> Option<&Cell<K, V>> {
+        let table = self.get_table();
+        if index >= table.size() {
+            return None;
+        }
+
+        let buckets = table.bucket_slice();
+        Some(Self::get_cell(buckets, index, table.size_mask))
+    }
+
+    pub(crate) fn get_cell_at_index_mut(&self, index: usize) -> Option<&'a mut Cell<K, V>> {
+        let table = self.get_table_mut();
+        if index >= table.size() {
+            return None;
+        }
+
+        let size_mask = table.size_mask;
+        let buckets = table.bucket_slice_mut();
+        Some(Self::get_cell_mut(buckets, index, size_mask))
+    }
+
     /// Gets a reference to a cell for the `index` from the `buckets`.
     #[inline]
-    fn get_cell(buckets: &[Bucket<K, V>], index: usize, size_mask: usize) -> &Cell<V> {
+    fn get_cell(buckets: &[Bucket<K, V>], index: usize, size_mask: usize) -> &Cell<K, V> {
         let bucket_index = Self::get_bucket_index(index, size_mask);
         let cell_index = Self::get_cell_index(index);
         &buckets[bucket_index].cells[cell_index]
@@ -676,7 +729,11 @@ where
     /// Gets a mutable reference mutable to a cell for the `index` from the
     /// `buckets`.
     #[inline]
-    fn get_cell_mut(buckets: &mut [Bucket<K, V>], index: usize, size_mask: usize) -> &mut Cell<V> {
+    fn get_cell_mut(
+        buckets: &mut [Bucket<K, V>],
+        index: usize,
+        size_mask: usize,
+    ) -> &mut Cell<K, V> {
         let bucket_index = Self::get_bucket_index(index, size_mask);
         let cell_index = Self::get_cell_index(index);
         &mut buckets[bucket_index].cells[cell_index]
@@ -735,16 +792,23 @@ where
     const fn get_cell_index(index: usize) -> usize {
         index & 3
     }
+}
 
-    /// Returns the null hash value.
-    #[inline]
-    const fn null_hash() -> HashedKey {
-        0 as HashedKey
-    }
+// This is safe to be send if the hasher and allocator are send, since the pointer
+// to the table is not thread-local and can be sent between threads.
+unsafe impl<K, V, H, A> Send for HashMap<K, V, H, A>
+where
+    H: BuildHasher + Send,
+    A: Allocator + Send,
+{
 }
 
 impl<K, V, H, A: Allocator> Drop for HashMap<K, V, H, A> {
     fn drop(&mut self) {
+        if !self.is_allocated() {
+            return;
+        }
+
         let table = self.get_table_mut();
 
         let bucket_ptr = table.buckets;
@@ -755,3 +819,119 @@ impl<K, V, H, A: Allocator> Drop for HashMap<K, V, H, A> {
         deallocate::<Table<K, V>, A>(&self.allocator, table_ptr, 1);
     }
 }
+
+impl<K, V, H, A> IntoIterator for HashMap<K, V, H, A>
+where
+    K: Eq + Hash + Clone,
+    V: Value,
+    H: BuildHasher + Default,
+    A: Allocator,
+{
+    type Item = (K, V);
+    type IntoIter = OwnedIter<K, V, H, A>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        OwnedIter::new(self)
+    }
+}
+
+impl<'a, K, V, H, A> IntoIterator for &'a HashMap<K, V, H, A>
+where
+    K: Eq + Hash + Clone,
+    V: Value,
+    H: BuildHasher + Default + 'a,
+    A: Allocator + 'a,
+{
+    type Item = (&'a K, &'a V);
+    type IntoIter = Iter<'a, K, V, H, A>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        self.iter()
+    }
+}
+
+/// Returns the null hash value.
+#[inline]
+pub(crate) const fn null_hash() -> HashedKey {
+    0_u64
+}
+
+/// The type used for hashed keys.
+pub(crate) type HashedKey = u64;
+
+/// Struct which stores a cell in a hash map. A cell is simply a hash (rather
+/// than the key iteself) and the value associated with the key for which the
+/// hash is associated.
+pub(crate) struct Cell<K, V> {
+    /// The hashed value of the cell. This adds 8 bytes of overhead to per key-value
+    /// pair for the cell, which is a lot, but it improves performance, so we
+    /// make the tradeoff.
+    /// FIXME: Reduce this overhead ...
+    hash: HashedKey,
+    // The key for the cell.
+    pub(crate) key: K,
+    /// The value assosciated with the hash.
+    pub(crate) value: V,
+}
+
+impl<K, V> Cell<K, V> {
+    /// Returns true if the cell is empty.
+    pub fn is_empty(&self) -> bool {
+        self.hash == null_hash()
+    }
+}
+
+/// Underlying table for a [HashMap]. This stores a pointer to the buckets and
+/// the mask for the number of cells which are stored in the table.
+struct Table<K, V> {
+    /// Pointer to the buckets for the table.
+    buckets: *mut Bucket<K, V>,
+    /// The mask for indexing into the buckets.
+    size_mask: usize,
+}
+
+impl<K, V> Table<K, V> {
+    /// Gets a mutable slice of the table buckets.
+    pub fn bucket_slice_mut(&mut self) -> &mut [Bucket<K, V>] {
+        unsafe { std::slice::from_raw_parts_mut(self.buckets, self.size()) }
+    }
+
+    /// Gets a slice of the table buckets.
+    pub fn bucket_slice(&self) -> &[Bucket<K, V>] {
+        unsafe { std::slice::from_raw_parts(self.buckets, self.size()) }
+    }
+
+    /// Returns the number of cells in the table.
+    pub fn size(&self) -> usize {
+        self.size_mask + 1
+    }
+}
+
+/// Struct which stores buckets of cells. Each bucket stores four cells
+/// and two delta values per cell. The first delta value for a cell provides
+/// the offset to the cell which is the start of the probe chain for the cell.
+/// The second delta value provides the offset to the next link in the probe
+/// chain once a search along the probe chain has started.
+struct Bucket<K, V> {
+    /// Delta values for the cells. The first 4 values are the delta values to
+    /// the start of the probe chain, and the second 4 values are the delta
+    /// values to the next probe in the chain, for each cell, respectively.
+    deltas: [u8; 8],
+    /// Cells for the bucket.
+    cells: [Cell<K, V>; 4],
+    /// Placeholder for the key
+    _key: std::marker::PhantomData<K>,
+}
+
+/// Defines the result of an insert into the [HashMap].
+enum InsertResult<V> {
+    /// The insertion found a cell for the key, replaced the old value with
+    /// a new one, and returned the old value.
+    Found(V),
+    /// The insertion was performed with a new key, so a new cell was filled.
+    NewInsert,
+    /// No new cell was found for the key withing the linear search range,
+    /// so we overflowed the max delta value and need to move the map to
+    /// a larger table.
+    Overflow(usize),
+}
